import json
import re
import logging
from bs4 import BeautifulSoup


# from https://isbn-checker.netlify.app
def valid_isbn(subject):
    "Check if the subject is a valid ISBN"

    isbn_regex = re.compile(
        r"^(?:ISBN(?:-1[03])?:? )?(?=[0-9X]{10}$|(?=(?:[0-9]+[- ]){3})[- 0-9X]{13}$|97[89][0-9]{10}$|(?=(?:[0-9]+[- ]){4})[- 0-9]{17}$)(?:97[89][- ]?)?[0-9]{1,5}[- ]?[0-9]+[- ]?[0-9]+[- ]?[0-9X]$"
    )

    # Check if the subject matches the ISBN pattern
    if isbn_regex.match(subject):
        chars = re.sub(r"[- ]|^ISBN(?:-1[03])?:?", "", subject)
        chars = list(chars)
        last = chars.pop()
        sum = 0
        check = 0

        if len(chars) == 9:
            chars.reverse()
            for i in range(len(chars)):
                sum += (i + 2) * int(chars[i])
            check = 11 - (sum % 11)
            if check == 10:
                check = "X"
            elif check == 11:
                check = "0"
        else:
            for i in range(len(chars)):
                sum += (i % 2 * 2 + 1) * int(chars[i])
            check = 10 - (sum % 10)
            if check == 10:
                check = "0"

        if str(check) == last:
            return True
        else:
            return False
    else:
        return False


# these are the currently supported identifier types that we can parse, along
# with their regex patterns
id_patterns = {
    # These come from https://gist.github.com/oscarmorrison/3744fa216dcfdb3d0bcb
    "isbn": [
        r"(?:ISBN(?:-10)?:?\ )?(?=[0-9X]{10}|(?=(?:[0-9]+[-\ ]){3})[-\ 0-9X]{13})[0-9]{1,5}[-\ ]?[0-9]+[-\ ]?[0-9]+[-\ ]?[0-9X]",
        r"(?:ISBN(?:-13)?:?\ )?(?=[0-9]{13}|(?=(?:[0-9]+[-\ ]){4})[-\ 0-9]{17})97[89][-\ ]?[0-9]{1,5}[-\ ]?[0-9]+[-\ ]?[0-9]+[-\ ]?[0-9]",
    ],
    # doi regexes taken from https://www.crossref.org/blog/dois-and-matching-regular-expressions/
    # listed in decreasing order of goodness. Not fully tested yet.
    "doi": [
        r"10.\d{4,9}\/[-._;()\/:A-Z0-9]+",
        r"10.1002\/[^\s]+",
        r"10.\d{4}\/\d+-\d+X?(\d+)\d+<[\d\w]+:[\d\w]*>\d+.\d+.\w+;\d",
        r"10.1021\/\w\w\d++",
        r"10.1207/[\w\d]+\&\d+_\d+",
    ],
}

# these can eliminate false positives
# TODO: remove duplication of validation logic and parsing logic
id_validators = {
    "isbn": valid_isbn,
}


def find_pdf_url(html_content) -> str | None:
    "Given HTML content, find an embedded link to a PDF."

    s = BeautifulSoup(html_content, "html.parser")

    # look for a dynamically loaded PDF
    script_element = s.find("script", string=re.compile("PDFObject.embed"))

    if script_element:
        match = re.search(r'PDFObject\.embed\("([^"]+)"', script_element.string)
        if match:
            logging.info("found dynamically loaded PDF")
            return match.group(1)

    # look for the "<embed>" element (scihub)
    embed_element = s.find("embed", {"id": "pdf", "type": "application/pdf"})

    if embed_element:
        direct_url = embed_element["src"]
        if isinstance(direct_url, list):
            direct_url = direct_url[0]
        if direct_url:
            logging.info("found embedded PDF")
            return direct_url

    # look for an iframe
    iframe = s.find("iframe", {"type": "application/pdf"})

    if iframe:
        logging.info(f"found iframe: {iframe}")
        direct_url = iframe.get("src")
        if isinstance(direct_url, list):
            direct_url = direct_url[0]
        if direct_url:
            logging.info("found iframe")
            return direct_url

    logging.info("No direct link to PDF found")
    return None


def parse_ids_from_text(
    s: str, id_types: list[str] | None = None
) -> list[dict[str, str]]:
    """
    Find all matches for the given id types in a string. If id_types isn't
    given, defaults to the types in id_patterns.
    """

    # we look for all ID patterns by default
    if id_types is None:
        id_types = list(id_patterns)

    seen = set()
    matches = []
    for id_type in id_types:
        for regex in id_patterns[id_type]:
            validator = id_validators.get(id_type)
            for match in re.findall(regex, s, re.IGNORECASE):
                valid_id = validator(match) if validator else True
                if match not in seen and valid_id:
                    matches.append({"id": match, "type": id_type})
                seen.add(match)
    return matches


def parse_file(path, id_types: list[str] | None = None):
    """
    Find all matches for the given id types in a file. If id_types isn't given,
    defaults to the types in id_patterns.
    """

    matches = []
    try:
        with open(path) as f:
            content = f.read()
        matches = parse_ids_from_text(content, id_types)
    except Exception as e:
        print(f"Error: {e}")

    return matches


def format_output(output: list[dict[str, str]], format: str = "raw") -> str:
    """
    Formats a list of dicts of ids and id types into a string according to the
    given format type. 'raw' formats ids by line, ignoring type. 'jsonl' and
    'csv' formats ids and types.
    """

    lines: list[str] = []
    if format == "raw":
        lines = [line["id"] for line in output]
    elif format == "jsonl":
        lines = [json.dumps(line) for line in output]
    elif format == "csv":
        lines = [f"{line['id']},{line['type']}" for line in output]
    else:
        raise Exception(f"invalid format {format}")
    return "\n".join(lines)



import argparse
import asyncio
import logging
import os
import sys

import aiohttp
from fetch import fetch_utils
from parse.parse import format_output, id_patterns, parse_file, parse_ids_from_text

supported_fetch_identifier_types = ["doi", "pmid", "url", "isbn"]


def parse_ids(args) -> str:
    output = None
    if hasattr(args, "path") and args.path:
        output = parse_file(args.path, args.match)
    else:
        # if a path isn't passed or is empty, read from stdin
        output = parse_ids_from_text(sys.stdin.read(), args.match)
    return format_output(output, args.format)


async def fetch(args):
    providers = args.providers
    id = args.query
    out = args.output

    headers = None
    if args.user_agent is not None:
        headers = {
            "User-Agent": args.user_agent,
        }

    async with aiohttp.ClientSession(headers=headers) as sess:
        pdf_content = await fetch_utils.fetch(
            sess,
            id,
            providers,
        )

    if pdf_content is None:
        return None
    path = os.path.join(out, fetch_utils.generate_name(pdf_content))
    fetch_utils.save(pdf_content, path)
    new_path = fetch_utils.rename(out, path)
    return new_path


async def run():
    name = "papers-dl"
    parser = argparse.ArgumentParser(
        prog=name,
        description="Download scientific papers from the command line",
    )

    from version import __version__

    parser.add_argument(
        "--version", "-V", action="version", version=f"{name} {__version__}"
    )

    parser.add_argument(
        "--verbose", "-v", action="store_true", help="increase verbosity"
    )

    subparsers = parser.add_subparsers()

    # FETCH
    parser_fetch = subparsers.add_parser(
        "fetch", help="try to download a paper with the given identifier"
    )

    parser_fetch.add_argument(
        "query",
        metavar="(DOI|PMID|URL)",
        type=str,
        help="the identifier to try to download",
    )

    parser_fetch.add_argument(
        "-o",
        "--output",
        metavar="path",
        help="optional output directory for downloaded papers",
        default=".",
        type=str,
    )

    parser_fetch.add_argument(
        "-p",
        "--providers",
        help="comma separated list of providers to try fetching from",
        default="all",
        type=str,
    )

    parser_fetch.add_argument(
        "-A",
        "--user-agent",
        help="",
        default=None,
        type=str,
    )

    # PARSE
    parser_parse = subparsers.add_parser(
        "parse", help="parse identifiers from a file or stdin"
    )
    parser_parse.add_argument(
        "-m",
        "--match",
        metavar="type",
        help="the type of identifier to search for",
        type=str,
        choices=id_patterns.keys(),
        action="append",
    )
    parser_parse.add_argument(
        "-p",
        "--path",
        help="the path of the file to parse",
        type=str,
    )
    parser_parse.add_argument(
        "-f",
        "--format",
        help="the output format for printing",
        metavar="fmt",
        default="raw",
        choices=["raw", "jsonl", "csv"],
        nargs="?",
    )

    parser_fetch.set_defaults(func=fetch)
    parser_parse.set_defaults(func=parse_ids)

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.ERROR)

    if hasattr(args, "func"):
        if asyncio.iscoroutinefunction(args.func):
            result = await args.func(args)
        else:
            result = args.func(args)

        if result:
            print(result)
        else:
            print("No papers found")
    else:
        parser.print_help()


def main():
    asyncio.run(run())


if __name__ == "__main__":
    asyncio.run(run())



import unittest

import aiohttp
import asyncio

from src.providers.scihub import get_available_scihub_urls


class TestSciHub(unittest.IsolatedAsyncioTestCase):
    async def test_scihub_up(self):
        """
        Test to verify that `scihub.now.sh` is available
        """
        urls = await get_available_scihub_urls()
        self.assertIsNotNone(urls, "Failed to find Sci-Hub domains")



import unittest

from src.parse.parse import find_pdf_url

test_cases = [
    (
        "scidb.html",
        "https://wbsg8v.xyz/d3/x/1719017408/134/i/scimag/80500000/80542000/10.1016/j.cub.2019.11.030.pdf~/Avtp6y0GwksOGlfLFy9d9Q/Parrots%20Voluntarily%20Help%20Each%20Other%20to%20Obtain%20Food%20Rewards%20--%20Brucks%2C%20D%C3%A9sir%C3%A9e%3B%20von%20Bayern%2C%20Auguste%20M_P_%20--%20Current%20Biology%2C%20%232%2C%2030%2C%20pages%20292-297_e5%2C%20--%2010_1016%2Fj_cub_2019_11_030%20--%20c28dc1242df6f931c29b9cd445a55597%20--%20Anna%E2%80%99s%20Archive.pdf",
    ),
    ("scihub.html", "https://sci.bban.top/pdf/10.1016/j.cub.2019.11.030.pdf"),
]


class TestSciDB(unittest.TestCase):
    def test_find_pdf_in_html(self):
        for file, expected_url in test_cases:
            with open(file, "rt") as f:
                html_content = f.read()
                pdf_url = find_pdf_url(html_content)
            self.assertEqual(pdf_url, expected_url)



import asyncio
import enum
import logging
import re
from urllib.parse import urljoin

import aiohttp
from bs4 import BeautifulSoup
from parse.parse import find_pdf_url

# URL-DIRECT - openly accessible paper
# URL-NON-DIRECT - pay-walled paper
# PMID - PubMed ID
# DOI - digital object identifier
IDClass = enum.Enum("identifier", ["URL-DIRECT", "URL-NON-DIRECT", "PMD", "DOI"])

DEFAULT_USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.3 Safari/605.1.15"


class IdentifierNotFoundError(Exception):
    pass


async def get_available_scihub_urls() -> list[str]:
    """
    Finds available Sci-Hub urls via https://sci-hub.now.sh/
    """

    # NOTE: This misses some valid URLs. Alternatively, we could parse
    # the HTML more finely by navigating the parsed DOM, instead of relying
    # on filtering. That might be more brittle in case the HTML changes.
    # Generally, we don't need to get all URLs.
    scihub_domain = re.compile(r"^http[s]*://sci.hub", flags=re.IGNORECASE)
    urls = []

    async with aiohttp.request("GET", "https://sci-hub.now.sh/") as res:
        s = BeautifulSoup(await res.text(), "html.parser")
    text_matches = s.find_all(
        "a",
        href=True,
        string=re.compile(scihub_domain),
    )
    href_matches = s.find_all(
        "a",
        re.compile(scihub_domain),
        href=True,
    )
    full_match_set = set(text_matches) | set(href_matches)
    for a in full_match_set:
        if "sci" in a or "sci" in a["href"]:
            urls.append(a["href"])
    return urls


async def get_direct_urls(
    session,
    identifier: str,
    base_urls: list[str] | None = None,
) -> list[str]:
    """
    Finds the direct source url for a given identifier.
    """
    if base_urls is None:
        base_urls = await get_available_scihub_urls()

    async def get_wrapper(url):
        try:
            return await session.get(url)
        except Exception as e:
            logging.error("error: %s" % e)
            return None

    if classify(identifier) == IDClass["URL-DIRECT"]:
        return [identifier]

    async with asyncio.TaskGroup() as tg:
        tasks = [
            tg.create_task(get_wrapper(urljoin(base_url, identifier)))
            for base_url in base_urls
        ]

    direct_urls = []
    # for item in zip(base_urls, tasks):
    for task in tasks:
        res = await task
        if res is None:
            continue
        path = find_pdf_url(await res.text())
        if isinstance(path, list):
            path = path[0]
        if isinstance(path, str) and path.startswith("//"):
            direct_urls.append("https:" + path)
        elif isinstance(path, str) and path.startswith("/"):
            print("res url type", type(res.url))
            direct_urls.append(urljoin(res.url.human_repr(), path))

    if not direct_urls:
        raise IdentifierNotFoundError

    logging.info(f"Found potential sources: {direct_urls}")
    return list(set(direct_urls))


def classify(identifier) -> IDClass:
    """
    Classify the type of identifier:
    url-direct - openly accessible paper
    url-non-direct - pay-walled paper
    pmid - PubMed ID
    doi - digital object identifier
    """
    if identifier.startswith("http") or identifier.startswith("https"):
        if identifier.endswith("pdf"):
            return IDClass["URL-DIRECT"]
        else:
            return IDClass["URL-NON-DIRECT"]
    elif identifier.isdigit():
        return IDClass["PMID"]
    else:
        return IDClass["DOI"]



import os
import unittest

from parse import parse

target_ids = ("doi", "pmid", "isbn", "issn", "url")


class TestParser(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_material_dir = "tests/identifiers"
        cls.valid_id_types = parse.id_patterns.keys()
        for id_type in target_ids:
            if id_type not in cls.valid_id_types:
                print(f"Skipping testing for {id_type} parsing")

    def test_parse_text(self):
        "Test to parse identifiers from a set of files."

        # NOTE: this test does not fail on false positive matches
        # for file in test_document_ids:
        for file in test_document_ids:
            print(f"testing {file}")
            with open(os.path.join(TestParser.test_material_dir, file)) as f:
                file_content = f.read()

            parsed_results = parse.parse_ids_from_text(file_content)

            # just include the matching id, not the type
            parsed_results = [result["id"] for result in parsed_results]

            expected_ids = []
            for type in test_document_ids[file]:
                if type in parse.id_patterns:
                    for id in test_document_ids[file][type]:
                        expected_ids.append(id)

            if not expected_ids:
                print("No expected IDs for this file")
                continue

            for expected_id in expected_ids:
                self.assertIn(
                    expected_id,
                    parsed_results,
                    f"ID {expected_id} not found in {file}",
                )


test_document_ids = {
    "ids.txt": {
        "url": ["https://www.cell.com/current-biology/fulltext/S0960-9822(19)31469-1"],
        "doi": ["10.1016/j.cub.2019.11.030"],
    },
    "bsp-tree.html": {
        "doi": ["10.1109/83.544569"],
        "issn": ["1057-7149", "1941-0042"],
    },
    "reyes-rendering.html": {
        "doi": ["10.1145/37402.37414"],
    },
    "superscalar-cisc.html": {
        "doi": ["10.1109/HPCA.2006.1598111"],
        "issn": ["1530-0897", "2378-203X"],
    },
    "b-tree-techniques.html": {
        "doi": ["10.1561/1900000028"],
        "url": ["http://dx.doi.org/10.1561/1900000028"],
        "isbn": ["978-1-60198-482-1", "978-1-60198-483-8"],
    },
    "real-time-rendering.html": {
        "url": ["https://doi.org/10.1201/9781315365459"],
        "isbn": ["9781315365459"],
    },
}



import unittest
import subprocess
import sys

test_paper_id = "10.1016/j.cub.2019.11.030"
test_paper_title = "Parrots Voluntarily Help Each Other to Obtain Food Rewards"


class TestCLI(unittest.TestCase):
    def test_parse_command_doi_csv(self):
        result = subprocess.run(
            [
                sys.executable,
                "src/papers_dl.py",
                "parse",
                "-m",
                "doi",
                "-p",
                "tests/identifiers/bsp-tree.html",
                "-f",
                "csv",
            ],
            capture_output=True,
            text=True,
        )
        self.assertIn("10.1109/83.544569,doi", result.stdout)

    def test_parse_command_doi_jsonl(self):
        result = subprocess.run(
            [
                sys.executable,
                "src/papers_dl.py",
                "parse",
                "-m",
                "doi",
                "-f",
                "jsonl",
                "-p",
                "tests/identifiers/bsp-tree.html",
            ],
            capture_output=True,
            text=True,
        )
        self.assertIn('{"id": "10.1109/83.544569", "type": "doi"}', result.stdout)

    def test_parse_command_isbn_raw(self):
        result = subprocess.run(
            [
                sys.executable,
                "src/papers_dl.py",
                "parse",
                "-m",
                "isbn",
                "-f",
                "raw",
                "-p",
                "tests/identifiers/b-tree-techniques.html",
            ],
            capture_output=True,
            text=True,
        )
        self.assertIn("978-1-60198-482-1", result.stdout)
        self.assertIn("978-1-60198-483-8", result.stdout)

    def test_parse_command_cli(self):
        args = [
            sys.executable,
            "src/papers_dl.py",
            "parse",
            "-f",
            "jsonl",
            "-m",
            "isbn",
        ]

        input_data = "978-1-60198-482-1 978-1-60198-483-8"

        result = subprocess.run(args, input=input_data, capture_output=True, text=True)
        self.assertIn('{"id": "978-1-60198-482-1", "type": "isbn"}', result.stdout)
        self.assertIn('{"id": "978-1-60198-483-8", "type": "isbn"}', result.stdout)



import asyncio
import hashlib
import json
import logging
import os
from typing import Iterable

import aiohttp
import pdf2doi
import providers.scidb as scidb
import providers.scihub as scihub

all_providers = [
    "scihub",
    "scidb",
]


def match_available_providers(
    providers, available_providers: Iterable[str] | None = None
) -> list[str]:
    "Find the providers that are included in available_providers"
    if not available_providers:
        available_providers = all_providers
    matching_providers = []
    for provider in providers:
        for available_provider in available_providers:
            # a user-supplied provider might be a substring of a supported
            # provider (sci-hub.ee instead of https://sci-hub.ee)
            if provider in available_provider:
                matching_providers.append(available_provider)
    return matching_providers


async def get_urls(session, identifier, providers):
    urls = []
    if providers == "all":
        urls.append(await scidb.get_url(session, identifier))
        urls.extend(await scihub.get_direct_urls(session, identifier))
        return urls

    providers = [provider.strip() for provider in providers.split(",")]
    logging.info(f"given providers: {providers}")

    matching_providers = match_available_providers(providers)
    logging.info(f"matching providers: {matching_providers}")
    for mp in matching_providers:
        if mp == "scihub":
            urls.extend(await scihub.get_direct_urls(session, identifier))
        if mp == "scidb":
            urls.append(await scidb.get_url(session, identifier))

    # if the catch-all "scihub" provider isn't given, we look for
    # specific Sci-Hub urls. if we find specific Sci-Hub URLs in the
    # user input, only search those
    if "scihub" not in providers:
        matching_scihub_urls = match_available_providers(
            providers, await scihub.get_available_scihub_urls()
        )
        logging.info(f"matching scihub urls: {matching_scihub_urls}")
        if len(matching_scihub_urls) > 0:
            urls.extend(
                await scihub.get_direct_urls(
                    session, identifier, base_urls=matching_scihub_urls
                )
            )

    return urls


async def fetch(session, identifier, providers):
    async def get_wrapper(url):
        try:
            return await session.get(url)
        except Exception as e:
            logging.error("error: %s" % e)
            return None

    urls = await get_urls(session, identifier, providers)

    logging.info("urls: %s" % urls)
    tasks = [get_wrapper(url) for url in urls if url]
    for item in zip(asyncio.as_completed(tasks), urls):
        res = await item[0]
        if res is None or res.content_type != "application/pdf":
            logging.info("couldn't find url at %s" % item[1])
            continue
        return await res.read()
    return None


def save(data, path):
    """
    Save a file give data and a path.
    """
    try:
        logging.info(f"Saving file to {path}")

        with open(path, "wb") as f:
            f.write(data)
    except Exception as e:
        logging.error(f"Failed to write to {path} {e}")
        raise e


def generate_name(content):
    "Generate unique filename for paper"

    pdf_hash = hashlib.md5(content).hexdigest()
    return f"{pdf_hash}" + ".pdf"


def rename(out_dir, path, name=None) -> str:
    """
    Renames a PDF to either the given name or its appropriate title, if
    possible. Adds the PDF extension. Returns the new path if renaming was
    successful, or the original path if not.
    """

    logging.info("Finding paper title")
    pdf2doi.config.set("verbose", False)

    try:
        if name is None:
            result_info = pdf2doi.pdf2doi(path)
            validation_info = json.loads(result_info["validation_info"])
            name = validation_info.get("title")

        if name:
            name += ".pdf"
            new_path = os.path.join(out_dir, name)
            os.rename(path, new_path)
            logging.info(f"File renamed to {new_path}")
            return new_path
        else:
            return path
    except Exception as e:
        logging.error(f"Couldn't get paper title from PDF at {path}: {e}")
        return path



from urllib.parse import urljoin

from parse.parse import find_pdf_url, parse_ids_from_text


async def get_url(session, identifier):
    base_url = "https://annas-archive.org/scidb/"

    is_doi = parse_ids_from_text(identifier, ["doi"])
    if is_doi:
        url = urljoin(base_url, identifier)
        res = await session.get(url)
        return find_pdf_url(await res.read())

    return None


doi_regexes = [
    r"10.\d{4,9}\/[-._;()\/:A-Z0-9]+",
    r"10.1002\/[^\s]+",
    r"10.\d{4}\/\d+-\d+X?(\d+)\d+<[\d\w]+:[\d\w]*>\d+.\d+.\w+;\d",
    r"10.1021\/\w\w\d++",
    r"10.1207/[\w\d]+\&\d+_\d+",
]
